__year__ = "2024"
__version__ = f"{__year__}.6.4"
__author__ = [
    "Aniq Fakhrul",
    "Ali Radzali"
]

BANNER = "Powerview.py v{} ({}) - by {}\n".format(__version__, __year__, ", ".join(__author__))
